 /*
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM.  Reproduction, modification and redistribution are prohibited.
 * VirtualGrid.js
 * (c) Copyright IBM Corporation 2011, 2014.  
 * U.S. Government Users Restricted Rights:  Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 */
 
 dojo.provide("com.ibm.rdm.web.grid.VirtualGrid");

(function () { 

dojo.require("dijit._Widget");
dojo.require("dijit._Templated");
dojo.require("dojo.i18n");
// to get default page size.
dojo.require("com.ibm.rdm.web.artifact.Constants");
dojo.require("com.ibm.rdm.web.notifications.ChangeNotifier");
dojo.require("com.ibm.rdm.web.commenting.EventTopics");
dojo.require("com.ibm.rdm.web.grid.fisheye.VirtualGridFisheye");
dojo.require("com.ibm.rdm.web.common.ErrorHandler");
dojo.require("com.ibm.rdm.web.common.MessageHandler");

dojo.requireLocalization("com.ibm.rdm.web.grid", "VirtualGrid");
// Provides virtual scrolling for the grid.  This class extends "Module" adding
// two page container DOM nodes to an absolutely positioned parent.  These two nodes
// display content loaded onto the Grid in the normal fashion by the superclass.
// 
// At present the DashboardTabPage is publishing an event on receipt of data from
// the view service.  We need this information to create a list of virtual pages mapped
// onto the view services pages.  e.g. if the view service says there are 200 total rows
// and 50 rows per page, then we need to build 4 virtual pages.  Each of these pages
// has the following key properties :
//
// top - this is the y coordinate of the top of the page
// height - the height of the page
// heightEstimated - flag indicating whether the height is an estimate.
//
// Once the first page is loaded, we measure it's height and remove it's height
// estimated flag.  For all pages that have an estimated height (the other 3 in the case)
// we estimate a height for them based on the height of rows in the first page and the number
// of rows they have.
// Once a height is either, known or estimated for each page, we sum them and size the parent node
// to be that many pixels high. 
//
// Finally, when the onscroll event fires, we use the scrollTop property to find the
// page that contains that position.  Publish a "GetPage" event (dojo.publish that is)
// which results in a call to the view service, we place the contents on one of the two
// page DOM nodes and order them in the DOM so that the lowest comes first. The wrapper
// is positioned absolutely at the lowest page numbers top coorodinate.  After each page
// is rendered it's height is measured and updated and the height of the grid body is also
// updated, so we should get more accurate as we go along, however scroll positions are adjusted
// where required to avoid the viewport contents jumping for the user.


// Fiddle with this to increase the sensitivity of the scrollbar
// This is a timeout after which the last
// received scroll event is processed. It prevents us hitting the server
// when the user has not actually paused to view a page. The flip side is that
// it affects the responsiveness as the user has to wait until this
// period has expired before a request to load the page is made.
var	SCROLL_THROTTLE_TIMER = 350,
	IMMEDIATE_PAGE_LOAD_THRESHOLD = 200,
	DEFAULT_PAGE_SIZE = com.ibm.rdm.web.artifact.Constants.DEFAULT_PAGE_SIZE,
	SCROLL_PREVIEW_HIDE_TIMEOUT = com.ibm.rdm.web.artifact.Constants.SCROLL_PREVIEW_HIDE_TIMEOUT;
	
var VirtualGridFisheye = com.ibm.rdm.web.grid.fisheye.VirtualGridFisheye;
var ChangeNotifier = com.ibm.rdm.web.notifications.ChangeNotifier;
var ErrorHandler = com.ibm.rdm.web.common.ErrorHandler;
var MessageHandler = com.ibm.rdm.web.common.MessageHandler;

dojo.declare("com.ibm.rdm.web.grid.VirtualGrid", com.ibm.rdm.web.grid.Module, {
  
  
	postMixInProperties: function () {
		this._messages = dojo.i18n.getLocalization("com.ibm.rdm.web.grid", "VirtualGrid");
	},	
  
	postCreate: function() {	
		dojo.addClass(this.domNode, "com-ibm-rdm-web-grid-VirtualGrid");
		this.totalRows = 0;
		this._samplePageIndex = 0;
		this.pageWrapper = dojo.create("div", {
			"class": "pageWrapper"
		}, this.domNode);
		
		this.loadingMsg = dojo.create("div", {
			"class": "loadingMsg",
			innerHTML: this._messages["loading"]
		}, this.scrollParent);
		
		var loadingGif = dojo.create("div", {
			"class": "loadingIcon"
		});
		
		dojo.place(loadingGif, this.loadingMsg, "first");
		
		// here's our two nodes that will actually hold page content
		// assume we are loading the first page until we know otherwise
		this.virtualPageOne = dojo.create("div", {
			"class": "vp",
			"page": "1"
		}, this.pageWrapper);
		this.virtualPageTwo = dojo.create("div", {
			"class": "vp"
		}, this.pageWrapper);
		// Disabling fisheye since structure is loaded differently now
		// See work item 81774
		this.hasFisheye = false;
		if (this.hasFisheye) {
			this.fisheye = new VirtualGridFisheye({
				grid: this,
				dataSource: this.pageModel
			});
		}
		
		// the superclass maintains all the connections it attaches in a
		// property called "dojoConnectList".  We need one of these for
		// each of our two pages.  When we determine that the page is to
		// be cleared we set this.dojoConnectList to point to one of these
		// and let the superclass do it's thing.
		this._pageOneConnectList = new com.ibm.rdm.web.grid.List();
		this._pageTwoConnectList = new com.ibm.rdm.web.grid.List();

		// Listen for changes to the page model, a change to the filter
		// or sort order results in us previously discarding the current
		// pages and scrolling to the top, so there is no further need
		// to react here.  
		this.connect(this.pageModel, "onChange", function(args) {
			
			if (!this.pages) {
				this._buildVirtualPages();
				if (this.pages.length > 0) {
					this.targetPage = this.pages[0];
				}
			} else {
				this._buildVirtualPages();
				if (!args || !args.gridContentLoad) {
					// "gridContentLoad" means that the code that triggered
					// the structural load is also going to load the grid
					// contents, so we needn't bother reconciling the grid
					// contents with the model.
					this._reconcilePages(args);
				} 
			}
			
			dojo.style(this.domNode, "height", this.pageModel.getTotalHeight()+"px");
		});
		
		ChangeNotifier.addListener({
			eventTopic: com.ibm.rdm.web.notifications.EventTopics.MODULE_COLUMN_CHANGE,
			callbackScope: this,
			callbackFunction: function(){
				// Update column for column addition or rearrangement.
				setTimeout(dojo.hitch(this, function() {
					var pageNum;
					// this.target has just been updated as a result of a column change, work out which of our two page DOM nodes
					// this.target currently is and refresh the other one if it is on display.
					if (this.target === this.virtualPageOne) {
						if (this._pageTwoRows && Object.getOwnPropertyNames(this._pageTwoRows).length > 0) {
							pageNum = this.virtualPageTwo.getAttribute("page");
							this._requestPage(this.pageModel.getPage(pageNum-1), this.virtualPageTwo);
						} 
					} else if (Object.getOwnPropertyNames(this._pageOneRows).length > 0) {
						pageNum = this.virtualPageOne.getAttribute("page");
						this._requestPage(this.pageModel.getPage(pageNum-1), this.virtualPageOne);
					}
				}), 100); // short timeout as _requestPage will change this.target, and this callback could happen
					      // before applyOrder is called.
			}
		});
		
		this.inherited(arguments);
	},
	
	_showLoadingMsg: function () {
		if (this._hideLoadingMsgAnim) {
			this._hideLoadingMsgAnim.stop();
		}
		dojo.style(this.loadingMsg, "display", "block");
		// we just forced a reflow...
		dojo.style(this.loadingMsg, {
			"marginLeft": ("-"+(this.loadingMsg.offsetWidth/2)+"px"),
			"marginTop": ("-"+(this.loadingMsg.offsetHeight/2)+"px")
		});
		
		this._showLoadingMsgAnim = dojo.fadeIn({node: this.loadingMsg});
		this._showLoadingMsgAnim.onEnd = dojo.hitch(this, function () {
			delete this._showLoadingMsgAnim;
		});
		
		this._showLoadingMsgAnim.play();
		
		
	},
	
	_hideLoadingMsg: function () {
		
		if (this._showLoadingMsgAnim) {
			this._showLoadingMsgAnim.stop();
		}
		
		this._hideLoadingMsgAnim = dojo.fadeOut({node: this.loadingMsg});
		this._hideLoadingMsgAnim.onEnd = dojo.hitch(this, function () {
			dojo.style(this.loadingMsg, "display", "none");
			delete this._hideLoadingMsgAnim;
		});
		
		this._hideLoadingMsgAnim.play();
	},
	
	onBodyScroll: function () {

		if (this._ignoreNextScrollEvent) {
			this._ignoreNextScrollEvent = false;
			return;
		}
		var scrollParent = this.scrollParent,
			scrollTop = scrollParent.scrollTop,
			page;
		
		
		if ((this._scrollTop === scrollTop) || (scrollTop === 0)) {
			// a horizontal scroll...
			return;
		}
		// save the scroll position in case the view is pushed out of DOM to cache.
		this._scrollTop = scrollTop;
		
		if (!this.pages) {
			return;
		}

		if (this._renderTimer) {
		   clearTimeout(this._renderTimer);
		   delete this._renderTimer;
		}
		
		if (!this._mouseUpListener) {
			// Other browsers do not fire the mouseup event when the mouse 
			// is over the scrollbar. 
			// see : http://code.google.com/p/chromium/issues/detail?id=14204
			//	     http://social.msdn.microsoft.com/Forums/en-US/netfxjscript/thread/3749b8a1-53ef-48fe-be81-b2df39d6154f/
			// however FF does and it is a useful cue to fetch the page without also waiting for the
			// scroll throttle timeout. 
			this._mouseUpListener = this.connect(document.body, "onmouseup", function () {
				if (this.hasFisheye) {
					this._hideThumbnailer();
					this.fisheye.hide();
				}
				this.disconnect(this._mouseUpListener);
				this._triggerPageLoad();
				delete this._mouseUpListener;
			});
		}		
		
		if (Math.abs(this._lastScrollTop - this._scrollTop) < IMMEDIATE_PAGE_LOAD_THRESHOLD) {
			this._triggerPageLoad();
		} else {
			this._renderTimer = setTimeout(dojo.hitch(this, this._triggerPageLoad), SCROLL_THROTTLE_TIMER);
		}
		this._lastScrollTop = this._scrollTop;
		
		if (this.hasFisheye) {
			this._showThumbnailer(scrollTop);
		}
	},
	
	
	_triggerPageLoad: function () {

		clearTimeout(this._renderTimer);
		
		
		if (this._inFlight) {
			// ensure only the last scroll event is queued for processing
			if (this.pendingScrollEvent) {
				dojo.disconnect(this.pendingScrollEvent);
				delete this.pendingScrollEvent;
			}
			
			this.pendingScrollEvent = dojo.connect(this, "onPageRendered", this, function () {
				dojo.disconnect(this.pendingScrollEvent);
				delete this.pendingScrollEvent;
				if (this.hasFisheye) {
					this._hideThumbnailer();
				}
				
				this._processScrollEvent(this.scrollParent.scrollTop);
				
			});
		} else {
			if (this.hasFisheye) {
				this._hideThumbnailer();
			}
			
			if (this.pages) {				
				this._processScrollEvent(this.scrollParent.scrollTop);
			}
			
		}
		
	},
	
	/// Overridden methods
	refreshRows: function(rowkeys) {
		if (rowkeys && dojo.isArray(rowkeys)) {
			var _allRows = this.allRows,
				rowIsOnPageOne;

			var key = rowkeys[0].key;
			var row = this._pageOneRows[key];
			
			if (row) {
				rowIsOnPageOne = true;
				this.allRows = this._pageOneRows;
			} else if (this._pageTwoRows){
				row = this._pageTwoRows[key];
				this.allRows = this._pageTwoRows;
			}

			this.inherited(arguments);
			setTimeout(dojo.hitch(this, this.measureAll));
		}
	},
	
	uninitialize: function () {
		this.hardClear();
		if (this.fisheye) {
			this.fisheye.destroyRecursive();
		}
		delete this.pages;
	},
	
	clear: function () {
		// override: no op..
	},
	
	hardClear: function () {
		this.clearAll();
		this._createContentNode();
		this._removeAllRows();
		this.setHeaderCheckbox(false);
	},
	
	clearAll: function () {
		this._clearPage(this.virtualPageOne);
		this._clearPage(this.virtualPageTwo);
		dojo.attr(this.virtualPageOne, "page", "1");
		this.target = this.virtualPageOne;
		dojo.style(this.pageWrapper, "top", "0px");
		this._samplePageIndex = 0;
		if (this.scrollParent) {
			this._ignoreNextScrollEvent = true;
			this.scrollParent.scrollTop = 0;	
			this.maxScrollPosRequested = false;
		}
		this.targetPage = this.pages && this.pages[0];
	},
	
	viewHasBeenRestored: function() {
		// summary:
		//		The grid has been restored from cache.  Restore the scroll position to
		//		the last known position.
		this.scrollParent.scrollTop = this._scrollTop;
		this._lastScrollTop = this._scrollTop;
		this._ignoreNextScrollEvent = true;
	},
	
	_clearPage: function (virtualPageNode, measurePage) {
		// summary:
		//		This function is invoked when the user changes the view,
		//		or when a page is reused for a different pages content.
		//		e.g. that page was displaying page 1, but the user has 
		//		scrolled to the bottom.
		//		The optional parameter indicates that before we clear
		//		out the page, update the list of pages (this.pages) with
		//		the height of the page.  Not required for a hard clear.
		var pageNum,
			page,
			tmp = this.allRows,
			keysToRelease = [];
		
		// first if we were asked to measure the height of the page,
		// try and do that now.	
		if (measurePage) {
			pageNum = dojo.attr(virtualPageNode, "page");
			if (pageNum) {
				page = this.pages[pageNum - 1];
			} else {
				return;
			}
		}
		
		dojo.query(".rdm-rich-text-editor", virtualPageNode).forEach(function (widgetNode) {
			var widget = dijit.getEnclosingWidget(widgetNode);
			if (widget && widget.ckEditor && typeof widget.suspend === "function") {
				widget.suspend();
			}
		});
		
				
		if (this._dragPointer && this._dragPointer.parentNode) {
			this._dragPointer.parentNode.removeChild(this._dragPointer);
		}
		
		if(dojo.query('.com-ibm-rdm-web-artifact-ModuleArtifactWidget.dragging')[0]) {
			// If the dragSource is removed from DOM the necessary drag events are not fired.
			var dragHandleNodes = dojo.query(".dragHandle.active", virtualPageNode);
			if(dragHandleNodes) {
				dragHandleNodes.forEach(function(node){
					node.className = "orphan-dragHandle";
					document.body.appendChild(node);	
				});
			}
		}
		
		// remove the child DOM nodes.
		dojo.empty(virtualPageNode);
		// get rid of the page attribute on the DOM node.
		dojo.removeAttr(virtualPageNode, "page");
		// here, we want to invoke the superclass cleanup functions
		// so we need to set this.allRows and this.dojoConnectList
		// to the correct objects for the page we are clearing
		if (virtualPageNode === this.virtualPageOne) {
			this.allRows = this._pageOneRows;
			this.dojoConnectList = this._pageOneConnectList;
		} else {
			this.allRows = this._pageTwoRows;
			this.dojoConnectList = this._pageTwoConnectList;
		}
		// now the allRows reference used in the base class points to the
		// correct set of resources for the page we want to empty. This is
		// a bit of a hack to reuse the row deletion functionality in the
		// base class
		
		// release the artifacts behind this page that are not selected.
		for (var key in this.allRows) {
			if (!this.selectedRows[key]) {
				keysToRelease.push(key);
			}
			
		}
		
		this.releaseArtifacts(keysToRelease);
			
		// remove all rows, but retain the selection...
		this._removeAllRows(true);
		this.setHeaderCheckbox(false);
		this._clearHandles();
		this.allRows = tmp;
		
		// 84416: We need to remember to make our internal
		// reference to the rows point at a new object.
		if (virtualPageNode === this.virtualPageOne) {
			this._pageOneRows = {};
		} else {
			this._pageTwoRows = {};
		}
		
	},
	
	setColumnWidth: function () {
		this.inherited(arguments);
		if (!this.pages) {
			return;
		}
		// setColumnWidth is called rather often by the grid, and since recalculating
		// the size of the page is expensive, and we only need to do it once we'll
		// set a short timeout to ignore the many invocations when the user changes
		// the columns.
		if (!this._pageHeightThrottle) {
			this._pageHeightThrottle = setTimeout(dojo.hitch(this, this._pageHeightChanged), 100);
		}
		
	},

	applyOrder: function () {

		var displayImages = this.gridmodel.isVisible(this.gridmodel.getPrimaryRowTypeId(), com.ibm.rdm.web.artifactgrid.ArtifactConstants.MODULE_CONTENT_COL);
						 
		this.pageModel.setDisplayImages(displayImages);

		this.content = this.target;
		this._hideLoadingMsg();
		this.allRows = {};	
		this.inherited(arguments);


		// this.allRows is now the page we just loaded
		if (this.target === this.virtualPageOne) {
			this._pageOneRows = this.allRows;
			this.dojoConnectList = this._pageOneConnectList;			
		} else {
			this._pageTwoRows = this.allRows;
			this.dojoConnectList = this._pageTwoConnectList;
		}

		setTimeout(dojo.hitch(this, function () {
			// TODO: deferred functions possibly mean this has to be
			// revised.  We happen to know that in Module, the deferred content
			// functions are invoked after a 0 timeout and usually have their
			// content available, as that content is coming from the ViewService
			// 
			var estimatedHeight;

			if (!this.pages) {
				this._buildVirtualPages();
				this.targetPage = this.pages[0];
				this.target = this.virtualPageOne;
			}
			
			if (this.targetPage) {
				// remove our estimated height, now we know how big this
				// page really is.
				estimatedHeight = this.targetPage.height;
				dojo.style(this.target, "height", "auto");
				this._measurePage();
				//
				if (this.adjustScroll) {
					this._ignoreNextScrollEvent = true;
					this.scrollParent.scrollTop += (this.targetPage.height - estimatedHeight);					
				}
				
				this.onPageRendered(this.targetPage);
			}
		}), 0);
	},
	
	measureAll: function () {
		// summary: 
		//		Measure the pages in DOM and update the page model.
		var vpOne = dojo.attr(this.virtualPageOne, "page"),
			vpTwo = dojo.attr(this.virtualPageTwo, "page");
			
		vpOne = vpOne && parseInt(vpOne, 10);
		vpTwo = vpTwo && parseInt(vpTwo, 10);
		if (vpOne) {
			this._measurePage(this.pages[vpOne-1], this.virtualPageOne);
		}
		
		if (vpTwo) {
			this._measurePage(this.pages[vpTwo-1], this.virtualPageTwo);
		}
	},
	
	_measurePage: function (page, pageNode) {
		page = page || this.targetPage;
		pageNode = pageNode || this.target;
		this._measureRequirementHeights(pageNode);
		this.pageModel.calculateHeightAndPositions();
		page.heightEstimated = false;
		dojo.style(this.domNode, "height", this.pageModel.getTotalHeight()+"px");
	},
	
	_measureRequirementHeights: function (node) {
		// measure the height of each requirement
		// and store in an array of heights for this page.
		
		// Note by SPC: In order to fix Defect 79825 I'm making use of the following
		// two variables to enable noticing the conditions that bring out the issue.
		var numRows = 0,
			lastHeight = 0;
		dojo.query(".resourceRow", node).forEach(dojo.hitch(this, function (node) {
			var uri = node.getAttribute("objectId");
			this.pageModel.setHeight(uri, node.offsetHeight);
			// Update variables needed to fix Defect 79825 ...
			numRows++;
			lastHeight = node.offsetHeight;
		}));
		// Here's the fix for Defect 79825.  When we notice that there's only a single
		// row on display AND that this single row's height is 0 we temporarily tweak
		// the CSS to force a repaint.
		if (numRows === 1 && lastHeight === 0) {
			var tempCssClass = "temporaryCssClassName";
			dojo.addClass(this.pageWrapper, tempCssClass);
			dojo.removeClass(this.pageWrapper, tempCssClass);
		}
	},
	
	onPageRendered: function () {
		// Stub fired when the virtual page has been populated
		this._inFlight = false;
		if (this.maxScrollPosRequested) {
			this._ignoreNextScrollEvent = true;
			this.scrollParent.scrollTop = this.domNode.offsetHeight;
		}

	},
	
	_createContentNode: function () {
		this.target = this.virtualPageOne;
	},
	
	_buildVirtualPages: function() {    
		this.pages = this.pageModel.getPages();  
	},

	_getVirtualPageShownOn: function (page) {
		// summary:
		//		Are either of the two virtual page nodes
		//		showing the content of page?
		var vpOne = dojo.attr(this.virtualPageOne, "page"),
			vpTwo = dojo.attr(this.virtualPageTwo, "page");
			
		if (vpOne == page.pageNum) {
			return this.virtualPageOne;
		} else if (vpTwo == page.pageNum) {
			return this.virtualPageTwo;
		}
		
		// implicit return of undefined
	},
	
	_getBestVirtualPageFor: function (page) {
		// summary:
		//		get the best of the two pages to put the content on
		//		e.g. if we are trying to load page 3 - but page 4 is loaded 
		//		on the next page, we might overlap with page 4, so put
		//		the content on the other page. 
		var vpOnePage = dojo.attr(this.virtualPageOne, "page"),
			vpTwoPage = dojo.attr(this.virtualPageTwo, "page");
		// sneaky type coercion...
		if (vpOnePage == page.pageNum+1 || vpOnePage == page.pageNum - 1) {
			return this.virtualPageTwo;
		} else if (vpTwoPage == page.pageNum+1 || vpTwoPage == page.pageNum - 1) {
			return this.virtualPageOne;
		} else {
			// doesn't matter
			return this.virtualPageOne;
		}
	},
	
	
	_processScrollEvent: function(scrollTop) {
		// summary:
		//		Get the page object that maps to the scroll
		//		position of scrollParent.scrollTop, 
	
		var page, 			// the page that matches the scroll pos
			pageDomNode, 	// the DOM node that the page either will be put on, or is on.
			spare,  		// the other of the two pages, i.e. the one not in use.
			nextPageAlreadyLoaded = false; 
			
		this.maxScrollPosRequested = ((scrollTop + this.scrollParent.offsetHeight) >= (this.domNode.offsetHeight - 2));
		page = this.pageModel.getPageForScrollPosition(scrollTop);
		if (!page) {
			page = this.pages[this.pages.length-1];
		}
		
		if (this._topItemGrabber) {
			clearTimeout(this._topItemGrabber);
			delete this._topItemGrabber;
		}
			
		// Remember the current top visible item of the viewport
		this._topItemGrabber = setTimeout(dojo.hitch(this, function() {
			this._lastAtTopOfViewport = this._getTopVisibleRow();			
		}), 250);
		
		pageDomNode = this._getVirtualPageShownOn(page);

		if (dojo.isFF && pageDomNode) {
			// FF appears to occasionally change it's mind about the height
			// of the first page, so ask again. Note this is not expensive
			// the browser already knows the answer to this, maybe we should do this
			// for all.
			page.height = pageDomNode.offsetHeight;
		}
		
		if (!pageDomNode) {
			pageDomNode = this._getBestVirtualPageFor(page);	
			spare = (pageDomNode === this.virtualPageOne) ? this.virtualPageTwo : this.virtualPageOne;
			nextPageAlreadyLoaded = (dojo.attr(spare, "page") == page.pageNum + 1);
			this._requestPage(page, pageDomNode, nextPageAlreadyLoaded);
			
			var con = dojo.connect(this, "onPageRendered", this, function(){
				dojo.disconnect(con);
				
				var targetNode;
				if (this.fisheye && !nextPageAlreadyLoaded && !this.maxScrollPosRequested) {
					var targetURI = this.fisheye.getTargetURI();
					targetNode = dojo.query("[objectId='"+targetURI+"']", this.domNode)[0];
					if (targetNode) {
						this._ignoreNextScrollEvent = true;
						this.scrollParent.scrollTop = this._getPositionRelativeToScrollParent(targetNode);
					} 
				}
				
				if (!this.pendingScrollEvent) {
					this._checkPageOverlap(page, spare);
					if (targetNode) {
						// it may look odd having to scroll into view twice,
						// but the first scroll into view may have caused an overlap
						// and checking the overlap also positions the container
						// relative to the body (that should proably be changed tho)
						// hence if the container moved, we need to scroll again.
						this._ignoreNextScrollEvent = true;
						this.scrollParent.scrollTop = this._getPositionRelativeToScrollParent(targetNode);
					} 
				}
				
			});
			
		} else {
			spare = (pageDomNode === this.virtualPageOne) ? this.virtualPageTwo : this.virtualPageOne;
			this._checkPageOverlap(page, spare);
		}
	},
	
	
	_getPositionRelativeToScrollParent: function (node) {
		// summary:
		//		Get the position of node relative to the scroll parent
		//		lets us set the scroll top for the scroll parent.
		//		as this is a private method and we know the structure
		//		of the DOM, just return the appropriate summed offsetTops
		return node.offsetTop + this.pageWrapper.offsetTop;
	},
	
	_checkPageOverlap: function(page, spare) {
		// summary:
		//		Does page overlap with another ?
		//		An overlap is where the virtual page does not fill all of the view port
		var scrollTop = this.scrollParent.scrollTop,
			overlapPage = this.pages[dojo.indexOf(this.pages, page)+1],
			otherPage = (spare === this.virtualPageOne) ? this.virtualPageTwo : this.virtualPageOne,
			bottomPage = this.pageModel.getPageForScrollPosition(scrollTop + this.scrollParent.offsetHeight),
			isOverlap;
		
		if (dojo.attr(spare, "page") == page.pageNum + 1) {	
			isOverlap = true;		
		} else if (page.pageNum != bottomPage.pageNum) {
			if (!page.isLast) {
				this._requestPage(overlapPage, spare);
				isOverlap = true;
			}
		} 
		
		// This code aggressively cleans up the unused page.  Supposing the user scrolls
		// from page 3 to 4, as soon as 3 is no longer in the viewport, this code would
		// clear the page meaning that if the user scrolled back to 3, we would need to load
		// it.  
		if (!isOverlap && dojo.attr(spare, "page") != page.pageNum - 1) {
			dojo.style(this.pageWrapper, "top", page.y+"px");
			this._clearPage(spare, true);
		}
	},
	
	
	_swapPages: function () {
		// Switch the two pages around so that they are in ascending
		// numerical order.  e.g. if we have pages 2 and 3 showing and 
		// the user scrolled upwards from 2 to reveal 1, then what
		// was page 3 becomes page 1, and needs to be placed
		// above page 2.
		var vpOne = dojo.attr(this.virtualPageOne, "page"),
			vpTwo = dojo.attr(this.virtualPageTwo, "page"),
			lowest;
			
		vpOne = vpOne && parseInt(vpOne, 10);
		vpTwo = vpTwo && parseInt(vpTwo, 10);
		
		if (vpOne && !vpTwo) {
			lowest = vpOne;
		} else if (vpTwo && !vpOne) {
			lowest = vpTwo;
		} else {
			lowest = Math.min(vpOne, vpTwo);
		}
		
		// Check if we actually need to swap the pages
		if(vpTwo !== null) {
			if(vpOne < vpTwo && this.virtualPageOne.nextElementSibling !== this.virtualPageTwo) {
				dojo.place(this.virtualPageOne, this.virtualPageTwo, "before");
			} else if(vpOne > vpTwo && this.virtualPageOne.nextElementSibling === this.virtualPageTwo) {
				dojo.place(this.virtualPageOne, this.virtualPageTwo, "after");
			}
		}
		return this.pages[lowest-1];
	},

	_requestPage: function(/*Object*/page, /*Node*/virtualPage, /*boolean*/adjustScroll) {
		// summary:
		//		Publish the "GetPage" event which causes the view service invocation.
		//		assigns this.target to the suppled "virtualPage" argument
		this.adjustScroll = adjustScroll;
		this.target = virtualPage;
		this._clearPage(virtualPage, true);
		this.targetPage = page;
		this.onGetPage(page.pageNum);
		this._inFlight = true;
		dojo.attr(virtualPage, "page", page.pageNum);
		if (!page.isLast) {
			dojo.style(virtualPage, "height", page.height + "px");
		}
		var pageAtTop = this._swapPages();
		dojo.style(this.pageWrapper, "top", pageAtTop.y+"px");
	},
	
	
	onGetPage: function (pageNum) {
		// stub fired when we wish to request a new page
		// 
		this._showLoadingMsg();
	},
	
	_pageHeightChanged: function () {
		// summary:
		//		Called when we know that the height of the page the user is looking
		//		at is changed.  This is probably due to a column resize.
		
		var targetURI, 
			targetNode,
			pageAtTop;
		
		if (this._pageHeightThrottle) {
			clearTimeout(this._pageHeightThrottle);
			delete this._pageHeightThrottle;
		}


		// what element is at the top of the viewport ?
		targetNode = this._lastAtTopOfViewport;
		
		// forget about the previously measured heights, they are now wrong
		// measure what we can.
		this.measureAll();
		// adjust the page we are looking at so it is in the correct place
		pageAtTop = this._swapPages();
		if (pageAtTop) {
			dojo.style(this.pageWrapper, "top", pageAtTop.y + "px");
		}
	},
	
	
	_showThumbnailer: function (scrollTop) {
		var page = this.pageModel.getPageForScrollPosition(scrollTop);
		if (this.hasFisheye) {
			this.fisheye.update(page);	
		}
		if (!page) {
			page = this.pages[this.pages.length-1];
		}
		if (!this._getVirtualPageShownOn(page)) {
			if (this.hasFisheye) {
				this.fisheye.show();	
			}
			this._hideThumbnailer();
		}
	},
	
	_hideThumbnailer: function () {
		if (this._hideThmbTimer) {
			clearTimeout(this._hideThmbTimer);
			delete this._hideThmbTimer;
		}
		
		if (this.hasFisheye) {
			this._hideThmbTimer = setTimeout(dojo.hitch(this, function () {
				this.fisheye.hide();
			}), SCROLL_PREVIEW_HIDE_TIMEOUT);
		}
	},

	requestOrReconcileArtifacts: function(resources, uris, callback){
		// summary:
		//		Invoked when this instance may or may not need artifacts
		//		uses artifacts in memory when possible (code lies in
		//		ResultSetGridView).
	},
	
	requestArtifacts: function (requiredUris, callback) {
		// summary:
		//		We would like to request the array of uris 
		//		so whatever contains the grid should load them.
	},
	
	releaseArtifacts: function (keys) {
	},

	
	_reconcilePages: function (/*Object?*/args) {
		this.pages = this.pageModel.getPages();
		var vpOne = dojo.attr(this.virtualPageOne, "page"),
			vpTwo = dojo.attr(this.virtualPageTwo, "page"),
			oldHeight,
			deferredArray = [];
			
		if (!this.pages || this.pages.length === 0) {
			this._removeAllRows();
			return;
		}
	
		vpOne = vpOne && parseInt(vpOne, 10);
		vpTwo = vpTwo && parseInt(vpTwo, 10);
		
		if (!vpTwo && vpOne && this.pages.length > 1) {
			// 83268:
			// The second "virtual page" DOM node does not currently
			// have a "page" attribute, presumably because it was not required
			// since the entire module could be contained on one page.
			// This is no longer the case - the this.pages[] is longer than 1,
			// and the first virtual page DOM node is in use.
			
			// first, assign the page number
			vpTwo = vpOne +1;
			// create an empty object (this is assigned to the property "allRows"
			// in the super class, it cannot be null or undefined).
			this._pageTwoRows = {};
			dojo.attr(this.virtualPageTwo, "page", vpTwo);
			// make sure they are in the correct order in the DOM
			this._swapPages();
		}
		
		// WI: 88694
		//
		// Where the user has added a new item which requires to be shown on 
		// a new page (e.g. the 51st row when the page size is 25) then the situation
		// could be that the we are currently displaying page 1 (1 - 25) and page 2 (25 - 50)
		// in our virtual page DOM nodes.  The model now has 3 pages, however since we
		// only reconcile the pages we are showing then the new artifact does not appear
		// since page 1 and page 2 both show the URIs that the model says they should.
		//
		// If we wait until the new item is scrolled into view, then because it is a "stub"
		// it can't be fetched from the view service.
		if (args && args.addItem && Array.isArray(args.uri) && args.uri.length > 0) {
			
			// if this is a new item...
		    var pageForNewItem = this.pageModel.getPageForURI(args.uri[0]);
		    
		    if (!pageForNewItem) {
		    	return;
		    }
		    
		    if (pageForNewItem.pageNum !== vpOne && pageForNewItem.pageNum !== vpTwo) {
		   		// and neither of our DOM nodes is showing the page the new item is on...
		   		
		   		// get the page that is either unused or is furthest in number away from the desired page.
		   		var pageDomNode = this._getBestVirtualPageFor(pageForNewItem);
		   		this._clearPage(pageDomNode);
				dojo.attr(pageDomNode, "page", pageForNewItem.pageNum);
				// make sure they are in the correct order in the DOM
				var pageAtTop = this._swapPages();
				dojo.style(this.pageWrapper, "top", pageAtTop.y+"px");
				// make sure these have the right values since we updated
				vpOne = dojo.attr(this.virtualPageOne, "page");
				vpTwo = dojo.attr(this.virtualPageTwo, "page");
				vpOne = vpOne && parseInt(vpOne, 10);
				vpTwo = vpTwo && parseInt(vpTwo, 10);
				
				// reconcile will now put the new row onto the page now we have
				// given it the correct number.
		   }
		}
		
		
		if (vpOne) {
			deferredArray.push(this._reconcilePage(this.pages[vpOne-1], this.virtualPageOne, this._pageOneRows, args));
			deferredArray[0].then(dojo.hitch(this, function (restoreFocusId) {
				this._measurePage(this.pages[vpOne-1], this.virtualPageOne);
				if (restoreFocusId) {
					this.moveFocusToCellOnRow(restoreFocusId);
				}
			}));
		}
		
		if (vpTwo) {
			deferredArray.push(this._reconcilePage(this.pages[vpTwo-1], this.virtualPageTwo, this._pageTwoRows, args));
			deferredArray[deferredArray.length - 1].then(dojo.hitch(this, function (restoreFocusId) {
				this._measurePage(this.pages[vpTwo-1], this.virtualPageTwo);
				if (restoreFocusId) {
					this.moveFocusToCellOnRow(restoreFocusId);
				}
			}));
		}
		
		if (deferredArray.length > 0) {
			// since the reconcile needs to update the target node, remember it here so that we
			// can reset it once reconciliation is done.  
			var tmpTarget = this.target;
			var deferredList = new dojo.DeferredList(deferredArray);
			deferredList.then(dojo.hitch(this, function(defResponses){
				this.target = tmpTarget;
				this._notifyGridReconciled();
			}));
		}
		else{
			this._notifyGridReconciled();
		}
	},
	
	_reconcilePage: function (page, domNode, rowList, /*boolean?*/args) {
		// summary:
		//		Compare the list of URIs actually on page with those that
		//		the page model thinks should be on it.  
		var urisOnPage,
			myUris = [],
			tmp = this.allRows,
			retainSelection = !args ? true : args.addItem;
		
		// make sure this.allRows points at the correct list for the page we are working on...
		this.allRows = rowList;
			
		if (page) {
			urisOnPage = page.uris;
		} else {
			urisOnPage = [];
		}
		
		for (var key in rowList) {
			myUris.push(key);
		}
		
		var toRemove = dojo.filter(myUris, function (uri) {
			return dojo.indexOf(urisOnPage, uri) === -1;
		});
		
		var toLoad = dojo.filter(urisOnPage, function (uri) {
			return dojo.indexOf(myUris, uri) === -1;
		});
		
		// get current focus node, so we can restore focus if any of the nodes
		// had it before we started shuffling them around.
		var focusElement = document.activeElement;
		// If set, the key of the row that focus should be restored to
		var restoreFocusId;

		// Note by SPC on 9/23/2013:
		// If the user has Escape'd out of creating a new row that's based on a stub
		// (e.g. one that was created via Create Similar) we need to make sure not
		// to lose focus as a result of removing the row.  We take care here to
		// set 'restoreFocusId' to an appropriate value when we detect that we're
		// cancel'ing out of the creation of such a row.
		if (args && !args.addItem && toRemove.length === 1) {
			var currentFocusRow = this.getFocusRow();
			if (currentFocusRow && currentFocusRow.key
				&& currentFocusRow.key === toRemove[0]
				) {
				var newFocusRowToUse = this.getFocusRowIfRowWasRemoved(currentFocusRow.key);
				if (newFocusRowToUse) {
					restoreFocusId = newFocusRowToUse.key;
				}
			}
		}

	//	console.debug("Reconcile, should remove ", toRemove);
	//	console.debug("Reconcile, should load ", toLoad);

		// only delete the rows to remove if there are none to load,
		// otherwise wait until the missing rows are loaded.
		if (!toLoad || toLoad.length === 0) {
			// if no args specified assume this is not a deletion, otherwise
			// use the addItem param to determine whether selection should
			// be retained.
			this.deleteRows(toRemove, true, retainSelection);
		}
		
		// if we have some rows in the DOM, but should not have, remove them now.
		if ((myUris.length === 0) && (dojo.query("table", domNode).length > 0)) {
			this._clearPage(domNode);
		}
		
		// ensure the order of the nodes is as it should be.
		var order = function () {
			// Only move nodes around if the order has changed.  This should be faster than moving them
			// if they don;t need to be moved, but also ensures that any ckeditors active on the rows
			// are not removed from DOM, which can kill them.
			var reorderNeeded,
				currentDisplayOrderURIs = [],
				pageSize = urisOnPage.length;
			dojo.query("table[objectId]", domNode).forEach(function(rowNode) {
				var objectId = dojo.attr(rowNode, "objectId");
				
				if (dojo.indexOf(urisOnPage, objectId) === -1) {
					// if this row is in the DOM, but should not be
					// according to the page model, remove it.
					rowNode.parentNode.removeChild(rowNode);
				} else {
					if (dojo.indexOf(currentDisplayOrderURIs, objectId) === -1) {
						currentDisplayOrderURIs.push(objectId);
					} else {
						// remove any duplicate rows
						rowNode.parentNode.removeChild(rowNode);
					}
				}
			});
			if (pageSize !== currentDisplayOrderURIs.length) {
				reorderNeeded = true;
			} else {
				for (var i = 0 ; i < pageSize ; i++) {
					if (urisOnPage[i] !== currentDisplayOrderURIs[i]) {
						// got at least one mismatch = need to reorder.
						reorderNeeded = true;
						break;
					}
				}
			}
			
			if (!reorderNeeded) {
				// avoid doing any reorder.
				return;
			}
			
			// Reorder the display nodes
			
			var fragment = document.createDocumentFragment();
			dojo.forEach(urisOnPage, function (uri) {
				var entry = rowList[uri];
				if (entry) {
					var node = entry.node;
					if (!restoreFocusId && node === focusElement) {
						restoreFocusId = uri;
					}
					fragment.appendChild(entry.node);
				} 
			});
			
			domNode.appendChild(fragment);
		};
		
		var deferred = new dojo.Deferred();
		if (toLoad && toLoad.length > 0) {
			var onLoad = dojo.hitch(this, function (resp) {
				// now we have obtained, via the view service, all the
				// things that were required to bring this page in sync
				// with the model, insert the rows onto the grid.
				
				
				
				var previousUri = this.pageModel.getPreviousURI(toLoad[0]);
				// is the previous URI on this page ? if not omit it, making
				// the row(s) appear at the top of the page.
				var uris = page.uris; 
				if (dojo.indexOf(uris, previousUri) === -1) {
					previousUri = null;
				}
				
				var newRows = [];
				for (var i=0; i<toLoad.length; i++) {
					newRows.push({key: toLoad[i], rowtype: "resourceRow"});
				}

				var insDeferred = this.preDisplayCallbackFunction && this.preDisplayCallbackFunction();
				if (insDeferred) {
					insDeferred.addCallback(dojo.hitch(this, function() {
						this.target = domNode;
						this.content = domNode;
						this.deleteRows(toRemove, true, retainSelection);
						this.insertRows(previousUri, newRows, true /* =noFocus */);
						order();
						deferred.callback(restoreFocusId);
						this.allRows = tmp;
					}));
				} else {
					this.target = domNode;
					this.content = domNode;
					this.insertRows(previousUri, newRows, true /* =noFocus */);
					order();
					deferred.callback(restoreFocusId);
					this.allRows = tmp;
				}
			});
			
			// If we are here as a result of an addExisting, then the command already
			// loaded the artifact for us, it will have passed it through as well so
			// need to request it again here...
			// The 'args' property 'resourceInGridModel' tells us that we don't need
			// to load the row.
			// We also don't need to load the view if the 'args' property
			// 'stubIdOfUserUnmodifiableRow' is passed.
			if (args && (args.resourceInGridModel || args.stubIdOfUserUnmodifiableRow)) {
				onLoad();
				this.allRows = tmp;
			} else {
				this.requestOrReconcileArtifacts((args && args.resources), dojo.clone(toLoad), onLoad);
			}
		} else {			 
			order();
			this.allRows = tmp;
			deferred.callback(restoreFocusId);
		}
		
		return deferred;
	},

	addToSelection: function(keys , /*optional*/ disableOnChange) {
		// summary:
		//		Override the base class method so that the "allRows" reference
		//		points to the correct rows for page 1 or page 2.  NOTE: This
		//		method is only designed to work for a single key.
		
		var firstKey = keys[0];
			
		if (keys.length !== 1) {
			this.inherited(arguments);
			return;
		}  

		if (this._pageOneRows && this._pageOneRows[firstKey]) {
			this.allRows = this._pageOneRows;
		}
		else if (this._pageTwoRows && this._pageTwoRows[firstKey]){
			this.allRows = this._pageTwoRows;
		} else {
//			ErrorHandler.debug("addToSelection key not found");
		}
	
		this.inherited(arguments);

	},
	

	_notifyGridReconciled: function() {
		// summary:
		//		Raise a notification to indicate that reconciliation of the grid has completed, so
		//		the rows in the grid should be stable.
		ChangeNotifier.notifyListeners({
			eventTopic: com.ibm.rdm.web.notifications.EventTopics.MODULE_GRID_RECONCILED
		});
	},
	
	getRowByKey: function(key) {
		var row;
		if (this._pageOneRows) {
			row = this._pageOneRows[key];
		}
		
		if (!row && this._pageTwoRows){
			row = this._pageTwoRows[key];
		}
		return row;
	},
	
	_insertRow: function(insertAfterHereKey, newKey, rowTypeId, functionsToCall) {
		// summary:
		//		Override the base class method so that we wire up the "allRows" reference
		//		to point to the correct rows for page 1 or page 2 and then call super().
		var found = false,
			row;
		if (insertAfterHereKey) {
			if (this._pageOneRows[insertAfterHereKey]) {
				this.allRows = this._pageOneRows;
				found = true;
			}
			else if (this._pageTwoRows && this._pageTwoRows[insertAfterHereKey]){
				this.allRows = this._pageTwoRows;
				found = true;
			}
			
			if (found) {
				row = this.inherited(arguments);
			}
			
		} else {
			this.allRows = this.target === this.virtualPageOne ? this._pageOneRows : this._pageTwoRows;
			row = this.inherited(arguments);
		}
		
		return row;
	 },
	 
	 isLoaded: function (key) {
		// summary:
		//		Returns whether or not the given key is loaded in the grid.  The key
		//		is loaded if it is currently displayed.
		
		if (this._pageOneRows && this._pageOneRows[key]) {
			return true;
		}
		if (this._pageTwoRows && this._pageTwoRows[key]) {
			return true;
		}
		return false;
	 },
	 
	 enableFisheye: function() {
		 if (this.hasFisheye) {
			 this.fisheye.enable();
		 }
	 },
	 
	 disableFisheye: function() {
		 if (this.hasFisheye) {
			 this.fisheye.disable();
		 }
	 },
	 
	 scrollIntoView: function (key, /* optional */ visibleCallback) {
		// summary:
		//		Scrolls the specified row into view, potentially loading
		//		the page it is on if required.
		
		// Firstly, if the node we are looking for is already loaded,
		// just scroll to it.
		var details = null;
		if (this._pageOneRows) {
			details = this._pageOneRows[key];
		}
		if (!details && this._pageTwoRows) {
			details = this._pageTwoRows[key];
		}
		
		if (details && details.node) {
			this._ignoreNextScrollEvent = true;
			dijit.scrollIntoView(details.node);
			if (visibleCallback) {
				visibleCallback();
			}
		} else {
			// it isn't, so find out from our module what page would contain
			// the key.
			var page = this.pageModel.getPageForURI(key);
			
			// TODO: move me somwhere common.
			// Wait until the page is rendered then scroll the artifact into view.
			// complication here is that if we scroll to a position that causes > 1 page to
			// rendered then this event will fire twice, make sure therefore to check the
			// page number.
			if (page) {
				var con = dojo.connect(this, "onPageRendered", this, function(loadedPage){
					if (loadedPage.pageNum === page.pageNum) {
						var targetNode = dojo.query("[objectId='" + key + "']", this.domNode)[0];
						if (targetNode) {
							this._ignoreNextScrollEvent = true;
							setTimeout(function () {
								dijit.scrollIntoView(targetNode);
							}, 0);
						}
						dojo.disconnect(con);
						if (visibleCallback) {
							visibleCallback();
						}
					}
				});
				
				this._ignoreNextScrollEvent = true;
				this.scrollParent.scrollTop = page.y + (page.height / 2);
				this._processScrollEvent(this.scrollParent.scrollTop);
			}
		}
	},
	
	getCellHtml: function(key, colId) {
		var row = null,
			results;
		if (this._pageOneRows) {
			row = this._pageOneRows[key];
		}
		if (row && row.node) {
			results = dojo.query("div."+colId, row.node); 
			if (results.length > 0) {
				return results[0].firstChild;
			} else {
//				ErrorHandler.debug("getCellHtml colId not found " + colId);
				return null;
			}
		}
		if (this._pageTwoRows) {
			row = this._pageTwoRows[key];
		}
		if (row && row.node) {
			results = dojo.query("div."+colId, row.node); 
			if (results.length > 0) {
				return results[0].firstChild;
			} else {
//				ErrorHandler.debug("getCellHtml colId not found " + colId);
				return null;
			}
		} 
//		ErrorHandler.debug("getCellHtml key not found " + key);
		return null;
	},
	
	
	selectBetweenIds: function (firstId, secondId) {

		var idxOfFirst = this.pageModel.indexOf(firstId),
			idxOfSecond = this.pageModel.indexOf(secondId),
			lowest = Math.min(idxOfFirst, idxOfSecond),
			highest = (lowest === idxOfFirst) ? idxOfSecond : idxOfFirst,
			selectionSize = (highest - lowest) + 1;


		if (selectionSize > com.ibm.rdm.web.artifact.Constants.MAX_MODULE_SELECTION_SIZE) {
			MessageHandler.showWarning(7669, [com.ibm.rdm.web.artifact.Constants.MAX_MODULE_SELECTION_SIZE]);
		} else {	
			this._selectAll(false);	
			var asList = this.pageModel.getArtifactsAsList(), 
				sublist = asList.slice(lowest, highest + 1);
				
			
			dojo.forEach(sublist, function(item){
				var objectId = item.uri;
				this.applyCheckRow(dojo.query("[" + com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID + "='" + objectId + "']", this.domNode)[0], objectId, "resourceRow");
			}, this);
		}
			
	},
	
	_getSelectionSize: function () {
		var key, selected = 0;
		
		for (key in this.selectedRows) {
			selected++;
		}
		
		return selected;
	},
	
	
	applyCheckRow: function(row, objectId, rowTypeId) {
		
		var numSelected = this._getSelectionSize(),
			alreadyInSelection = this.selectedRows[objectId],
			maxSelectionSize = com.ibm.rdm.web.artifact.Constants.MAX_MODULE_SELECTION_SIZE;
		
		// Confusingly, this.selectedRows already includes the selection if we are selecting
		// via the keyboard, this is probably incorrect but it is not a big deal to work this 
		// out as below.
		if ( (alreadyInSelection && (numSelected > maxSelectionSize)) || 
			(!alreadyInSelection && (numSelected === maxSelectionSize))) {
			// disallow the selection...
//			console.debug("Not allowed to select...");
			dojo.query("."+com.ibm.rdm.web.grid.Module.ClassNames.MAIN_SELECT, row).forEach(function (item) {item.checked = false; });
			MessageHandler.showWarning(7669, [com.ibm.rdm.web.artifact.Constants.MAX_MODULE_SELECTION_SIZE]);
			delete this.selectedRows[objectId];
		} else {
			if (this.gridmodel.isSelectable(this.gridmodel.getPrimaryRowTypeId()) || dojo.hasClass(row, "underEdit" )) {
				this.selectedRows[objectId] = {
					rowTypeId: "resourceRow",
					uri: objectId
				};
				
				if (row) {
					this.inherited(arguments, [row, objectId, rowTypeId, true]);
				}
			}
			
		}
		
	},
	
	onKeyUpDownAtGridEnd: function(goUp, currentRowKey) {
		// summary:
		//		Override for method on Module.js
		//		Called when there is an key down event on the bottom row in the grid.  Indicates
		//		that a move to a new page in the virtual grid may be necessary.
		//		If necessary, calls focusCorrespondingCellInNewRow() on the underlying grid to
		//		change focus once the new row is available (may be after the page has been loaded).
		// goUp:
		//		Boolean indicating whether we are dealing with a keyUp at the start of the page
		//		or a keydown at the end.
		// currentRowInfo:
		//		Row information containing, amongst other fields, a key (URI) field indicating the
		//		current focus row.
		//		
		var nextRowKey = goUp ? this.pageModel.getPreviousURI(currentRowKey) : this.pageModel.getNextURI(currentRowKey);
		if (nextRowKey) {
			this.scrollIntoView(nextRowKey, dojo.hitch(this, this._focusRowOrCell, nextRowKey));
		}
	},
	
	_focusRowOrCell: function(targetRowKey) {
		// summary:
		//		Moves focus to the row, or to a cell on the row, depending on whether
		//		the current focus is on a cell or on the whole row.
		// targetRowKey:
		//		The key (URI) of the row on which to focus.
		var targetRow = this.getRowByKey(targetRowKey);
		this.focusCorrespondingCellInNewRow(targetRow.node);
	},
	
	setFocusForPageUpDown: function(goUp, currentRowKey) {
		// summary:
		//		Override for the corresponding method on Module.js
		//		Find the row that is approximately one viewport distance up/down from the given row.
		// goUp:
		//		Indicates whether we are doing a page up or page down from the current position.
		// currentRowKey:
		//		Key for the current row position.
		
		var viewportHeight = this.scrollParent.offsetHeight;
		var targetRowKey = goUp ? 
							this.pageModel.getEarlierRowAtOffset(currentRowKey, viewportHeight) : 
							this.pageModel.getLaterRowAtOffset(currentRowKey, viewportHeight);
		if (targetRowKey) {
			this.scrollIntoView(targetRowKey, dojo.hitch(this, this._focusRowOrCell, targetRowKey));
		}
	},
	
	setFocusForCtrlHomeEnd: function(goHome) {
		// summary:
		//		Override for the corresponding method on Module.js
		//		Virtual grid handler for the Ctrl-Home and Ctrl-End keys.  Moves focus to the
		//		first or last row on the virtual grid.
		// goHome:
		//		Boolean.  Indicates whether we are dealing with the home key or the end key.
		var pages = this.pageModel.getPages();
		if (pages.length === 0) {
			// empty document - nothing to do
			return;
		}
		var	targetPageIndex = goHome ? 0 : (pages.length-1),
			targetPage = pages[targetPageIndex],
			targetRowKey = targetPage.uris[goHome ? 0 : (targetPage.uris.length-1)];
		if (targetRowKey) {
			this.scrollIntoView(targetRowKey, dojo.hitch(this, this._focusRowOrCell, targetRowKey));
		}
	},
	
	handleSelectAllKey: function() {
		// summary:
		//		Handle keypress to select all items in the grid.  Override of the corresponding base class
		//		method.  In the virtual grid, select all is not permitted, so this method does nothing.
		// returns:
		//		Always returns true, to indicate that the key has not been handled and so can be passed to
		//		handlers further down the chain.
		return true;
	},
	
	_findNodeByObjectId: function (/*string*/objectId) {
		// summary:
		//		Return the DOM node that has an objectid attribute matching the parameter.
		//
		//      Override this here so that we look in the current target page DOM node.
		return dojo.query("table["+com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID+"='"+objectId+"']", this.target)[0];
	},
		
	_getTopVisibleRow: function() {
		// summary:
		//		Get the top visible row from the viewport. 
	    var scrollTop = this.scrollParent.scrollTop,
	    	found = false,
	    	topRow = null;

		dojo.query("table", this.domNode).forEach(function(row) {
			if (!found) {
				var rowTop = row.offsetTop + row.offsetHeight;
				if (this.pageWrapper) {
					rowTop += this.pageWrapper.offsetTop;
				}
				if (rowTop > scrollTop) {
					found = true;
					topRow = row;
				}
			}
		}, this);
		return topRow;
	},
	
	scrollNodePosition: function(node, isTopItem) {
		// summary:
		//		Restore the scroll position by scrolling the given row into the viewport. 
		//		If the node is a top item scroll it to the top of the viewport.

		// Get the position of the node relative to the scroll parent
		var rowPos = this._getPositionRelativeToScrollParent(node);
		if (!isTopItem) {
			// If the node is a bottom item, calculate the top row position base 
			// on the bottom item. And use the top row position for the scroll.
			var topRow = this._getTopVisibleRow();
			if (topRow) {
				var delta = rowPos - topRow.offsetTop;
				rowPos = rowPos - delta;
			}
		}
		this.scrollParent.scrollTop = rowPos;
		this._lastScrollTop = rowPos;
		this._ignoreNextScrollEvent = true;
		this._processScrollEvent(this.scrollParent.scrollTop);
	},
	
	getTopVisibleItemFromViewport: function() {
		// summary:
		//		Get the top visible row item from the viewport. 
		var key = null;
		var row = this._lastAtTopOfViewport;
		if (row) {
			key = dojo.attr(row, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
		}
		return key;
	},
	
	getBottomVisibleItemFromViewport: function() {
		// summary:
		//		Get the bottom visible row item from the viewport.
		var found = false,
			viewportHeight = this.scrollParent.scrollTop + this.scrollParent.offsetHeight,
			key = null;
		
		dojo.query("table", this.domNode).forEach(function(row) {
			if (!found) {
				if (row.offsetTop >= viewportHeight) {
					found = true;
				}
				else {
					key = dojo.attr(row, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
				}
			}
		});
		return key;
	}
});

} ());
